import sys

from Accuinsight.modeler.entities._modeler_object import _ModelerObject
from Accuinsight.modeler.protos.service_pb2 import Param as ProtoParam


class Param(_ModelerObject):
    """
    Parameter object.
    """

    def __init__(self, key, value):
        if "pyspark.ml" in sys.modules:
            import pyspark.ml.param
            if isinstance(key, pyspark.ml.param.Param):
                key = key.name
                value = str(value)
        self._key = key
        self._value = value

    @property
    def key(self):
        """String key corresponding to the parameter name."""
        return self._key

    @property
    def value(self):
        """String value of the parameter."""
        return self._value

    def to_proto(self):
        param = ProtoParam()
        param.key = self.key
        param.value = self.value
        return param

    @classmethod
    def from_proto(cls, proto):
        return cls(proto.key, proto.value)



import logging
import re
import os
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

logging.basicConfig(level=logging.DEBUG)


class AbcNotifier:
    def completed_message(self, **kwargs):
        pass


class SlackApp(AbcNotifier):
    def __init__(self, token, channel_id):
        self.token = token
        self.channel_id = channel_id
        self.client = WebClient(token=self.token)

    def send_message(self, message='Job is done.'):
        try:
            response = self.client.chat_postMessage(
                channel=self.channel_id,
                text=message
            )
        except SlackApiError as e:
            assert e.response["error"]

    def completed_message(self, **kwargs):
        self.send_message(message=kwargs.get('message'))


def load_model(run_name):
    k = []
    best_model_path_dict = dict()

    model_type = run_name.split('-')[0]
    t = run_name.split('-')[1].split('_')[0]
    filename = model_type + '-' + t[0:8] + '-' + t[8:12] + '-' + t[12:16] + '-' + t[16:20] + '-' +  t[20:]
    regex = re.compile(filename)

    common_path = '/home/work/runs/best-model'

    for subdir, dirs, files in os.walk(common_path):
        for filename in files:
            if regex.search(filename):
                k.append(os.path.join(common_path, filename))

    if 'json' in k[0]:
        best_model_path_dict['json'] = k[0]
        best_model_path_dict['h5'] = k[1]
    elif 'h5' in k[0]:
        best_model_path_dict['json'] = k[1]
        best_model_path_dict['h5'] = k[0]
    else:
        best_model_path_dict['joblib'] = k[0]

    # load model trained using keras
    if model_type == 'keras':
        from keras.models import model_from_json
        # load json and create model
        json_file = open(best_model_path_dict['json'], 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        loaded_model = model_from_json(loaded_model_json)

        # load weights into new model
        loaded_model.load_weights(best_model_path_dict['h5'])

    # load model trained using tensorflow
    elif model_type == 'tf.keras':
        from tensorflow.keras.models import model_from_json
        # load json and create model
        json_file = open(best_model_path_dict['json'], 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        loaded_model = model_from_json(loaded_model_json)

        # load weights into new model
        loaded_model.load_weights(best_model_path_dict['h5'])

    else:
        import joblib
        loaded_model = joblib.load(best_model_path_dict['joblib'])

    return loaded_model



import json
from Accuinsight.modeler.core.Run.RunInfo.RunInfo import get_current_run
from Accuinsight.modeler.core.LcConst.LcConst import RUN_BASE_PATH, RUN_PREFIX_PATH
from Accuinsight.modeler.core.LcConst.LcConst import ALL_MODEL_PARAMS, SELECTED_METRICS, SELECTED_PARAMS, RUN_OBJ_MODEL_JSON_PATH
from Accuinsight.modeler.protos.life_cycle_pb2 import LcParam, LcMetric
from Accuinsight.modeler.core.Run.ParseRun import Parse_Helper


# load json data
def _load_current_run():
    run_results = {'json':'', 'visual':'', 'run_mata': ''}

    run_info_json = get_current_run()

    base_path = run_info_json[RUN_BASE_PATH]
    prefix_path = run_info_json[RUN_PREFIX_PATH]
    data_path = run_info_json[RUN_OBJ_MODEL_JSON_PATH]

    data_path = ''.join([base_path, prefix_path, data_path])

    with open(data_path) as json_file:
        json_data = json.load(json_file)

    run_results['json'] = json_data
    run_results['run_mata'] = run_info_json

    return run_results


def _makeLcParam(selected_params, all_model_params):
    grpc_params = []

    for param_key in selected_params:

        param = LcParam(
            key=param_key,
            value=str(all_model_params[param_key])
        )
        grpc_params.append(param)
    return grpc_params


def _append_metrics(metrics, **kwargs):
    values = []
    timestamp = []
    steps = []
    count = 0
    data = kwargs['data']
    key = kwargs['key']

    for v in data:
        values.append(str(v))
        timestamp.append(str(count))
        steps.append(str(count))
        count += 1
    metrics.append({'key': key,
                    'values': values,
                    'timestamp': timestamp,
                    'steps': steps})
    return metrics


def _parse_true_predict(metric_data, metrics):
    true_y = []
    predicted_y = []
    true_y_values = metric_data['True_y']
    predicted_y_values = metric_data['Predicted_y']

    # check if nested list
    is_nested_true = any(isinstance(elem, list) for elem in true_y_values)
    is_nested_predicted = any(isinstance(elem, list) for elem in predicted_y_values)

    if is_nested_true:
        for elem in true_y_values:
            true_y.append(elem[0])
    else:
        true_y = true_y_values

    if is_nested_predicted:
        for elem in predicted_y_values:
            predicted_y.append(elem[0])
    else:
        predicted_y = predicted_y_values

    metrics = _append_metrics(metrics, data=true_y, key='true_y')
    metrics = _append_metrics(metrics, data=predicted_y, key='predicted_y')
    return metrics


# parse parameters
def parse_run_parameters(json_data):
    # read ~/runs/results-XGBClassifier/model-info-json/*.json
    # return parameters

    param_data = json_data['json']
    if SELECTED_PARAMS in param_data:
        selected_params = param_data[SELECTED_PARAMS]

    all_model_params = param_data[ALL_MODEL_PARAMS]

    grpc_params = _makeLcParam(selected_params=selected_params, all_model_params=all_model_params)

    data_version = Parse_Helper.get_data_version(param_data)
    grpc_params.append(data_version)

    return grpc_params


# parse metrics
def parse_run_metrics(json_data):
    # read ~/runs/results-XGBClassifier/model-info-json/*.json
    # get "selected_metrics" field
    # return metrics

    metric_data = json_data['json']
    grpc_metrics = []

    # selected_metrics
    selected_metrics = metric_data[SELECTED_METRICS]
    metric_keys = selected_metrics.keys()

    result_dict = {'metrics': []}

    count = 0
    for key in metric_keys:
        timestamp = []
        steps = []
        values = []

        values.append(str(selected_metrics[key]))
        timestamp.append(str(count))
        steps.append(str(count))
        count += 1

        result_dict['metrics'].append({'key': key,
                                       'values': values,
                                       'timestamp': timestamp,
                                       'steps': steps})

    metrics = result_dict['metrics']
    
    metrics = _parse_true_predict(metric_data, metrics)

    for item in metrics:
        metric = LcMetric(
                key=item['key'],
                values=item['values'],
                timestamp=item['timestamp'],
                steps=item['steps']
            )
        grpc_metrics.append(metric)

    return grpc_metrics


# parse metrics
def _append_visual_data(visual_data, key):
    result_dict = {'metrics': []}

    count = 0
    timestamp = []
    steps = []
    values = []

    for k, v in visual_data.items():
        timestamp.append(str(count))
        count += 1

        steps.append(k)

        values.append(v)

    result_dict['metrics'].append({'key': key,
                                   'values': values,
                                   'timestamp': timestamp,
                                   'steps': steps})
    return result_dict


def _make_steps(steps_value):
    steps_list = []
    for value in steps_value:
        steps_list.append(str(value))
    return steps_list


def parse_run_result():
    result_dict = {'metrics': '', 'params': ''}
    result_data = _load_current_run()

    grpc_params = parse_run_parameters(json_data=result_data)
    grpc_metrics = parse_run_metrics(json_data=result_data)

    result_dict['params'] = grpc_params
    result_dict['metrics'] = grpc_metrics
    for i in range(len(grpc_params)):
        if grpc_params[i].key == 'data_version':
            result_dict['artifact'] = {}
            result_dict['artifact']['name'] = grpc_params[i].value
            result_dict['artifact']['version'] = ""
            del grpc_params[i]
            break

    return result_dict




import datetime, json

from Accuinsight.modeler.entities.monitoring_deploy_log import DeployLog
from Accuinsight.modeler.entities.monitoring_deploy_alarm import DeployAlarm
from Accuinsight.modeler.clients.modeler_api import DeployLogRestApi
from Accuinsight.modeler.core.MonitoringConst.deploy import DeployConst
from Accuinsight.modeler.utils.os_getenv import get_os_env


class AddDeployLog:
    """
        Object for adding deploy log.
    """

    def __init__(self):
        env_value = get_os_env('ENV')

        self.deploy_log = DeployLog()
        self.deploy_alarm = DeployAlarm()
        self.deploy_log_api = DeployLogRestApi(env_value[DeployConst.BACK_END_API_URL],
                                               env_value[DeployConst.BACK_END_API_PORT],
                                               env_value[DeployConst.BACK_END_API_URI])
        self.swagger = False

    def set_request(self, request):
        self.deploy_log.start_time = datetime.datetime.now().timestamp()
        self.deploy_log.req_method = request.method
        self.deploy_log.req_url = request.url

        req_data = list()
        if request.data:
            req_data.append(request.data.decode('utf-8'))

        if request.files:
            req_data.append(dict(request.files))

        if request.form:
            req_data.append(dict(request.form))

        if len(req_data) == 1:
            req_str = str(req_data.pop())
        elif len(req_data) == 0:
            req_str = str()
        else:
            req_str = str(req_data)

        try:
            self.deploy_log.req_body = json.loads(req_str)
        except json.decoder.JSONDecodeError:
            self.deploy_log.req_body = req_str

        if "swagger" in request.url:
            self.swagger = True
        else:
            self.swagger = False

    def set_response(self, response):
        response.direct_passthrough = False

        self.deploy_log.end_time = datetime.datetime.now().timestamp()
        self.deploy_log.status_code = str(response.status_code)
        try:
            res_data = response.get_data().decode()
        except UnicodeDecodeError:
            res_data = response.get_data()

        try:
            self.deploy_log.response_data = json.loads(res_data)
        except json.decoder.JSONDecodeError:
            self.deploy_log.response_data = str(res_data)

    def add_log(self, messages, notifiers):
        if not self.swagger:
            if messages is not None:
                self.deploy_alarm.timestamp = datetime.datetime.now().timestamp()

                if 'slack' in notifiers.keys():
                    self.deploy_alarm.notifiers['slack'] = notifiers['slack']['hook_url']

                if 'mail' in notifiers.keys():
                    self.deploy_alarm.notifiers['mail'] = notifiers['mail']['address']

                for message in messages:
                    self.deploy_alarm.message = message
                    self.deploy_log_api.call_rest_api('POST', self.deploy_alarm.get_alarm_param(), "alarm")

            self.deploy_log_api.call_rest_api('POST', self.deploy_log.get_logging_param())

    def get_log_info(self):
        try:
            summary_data = json.loads(self.deploy_log_api.call_rest_api('GET', None, 'log_summary'))['data']
        except (TypeError, KeyError):
            summary_data = {'totalCall': 0, 'totalSuccessCall': 0}

        log_info = dict()
        log_info['total_call'] = summary_data['totalCall']
        log_info['total_success_call'] = summary_data['totalSuccessCall']
        log_info['latest_log'] = {
            'start_time': datetime.datetime.fromtimestamp(self.deploy_log.start_time),
            'end_time': datetime.datetime.fromtimestamp(self.deploy_log.end_time),
            'duration': round(self.deploy_log.end_time - self.deploy_log.start_time, 3),
            'url': self.deploy_log.req_url,
            'request_method': self.deploy_log.req_method,
            'request_body': self.deploy_log.req_body,
            'status_code': int(self.deploy_log.status_code),
            'response_data': self.deploy_log.response_data
        }

        return log_info



import os
import itertools

from Accuinsight.modeler.core import get, metricsLightgbm
from Accuinsight.modeler.core.LcConst.LcConst import RUN_ROOT_PATH
from Accuinsight.modeler.core.Run.RunInfo.RunInfo import get_or_create_run_directory


def get_file_path(model_name, root_path = RUN_ROOT_PATH, **kwargs):
    """
    ML 모델 저장을 위한 `add_experiment()`, 또는 DL 모델 저장을 위한 `autolog()` 실행시 output 저장을 위한 경로 생성 함수

    :model_name: 실행된 모델의 이름
    :root_path:
    :kwargs: DL framework를 구분하기 위한 변수 입력, (eg. usedFramework = 'keras')

    """
    
    # 기존 path 저장 - path 변경으로인한 에러 방지
    original_path = os.getcwd()
    
    # data 저장 상위 폴더
    home = get_or_create_run_directory()
    os.chdir(home)
    
    try:
        # DL frameworks(keras/tf/pytorch)
        a_dict = {}
        for key, value in kwargs.items():
            a_dict[key]= value
    
        prefix_path = 'results'+ '-' + list(a_dict.values())[0] + '/'
    
    except IndexError:
        # ML
        prefix_path = 'results'+ '-' + get.model_type(model_name) + '/'

    if not os.path.isdir(prefix_path):
        try:
            os.mkdir(prefix_path)
        except:
            pass
    
    os.chdir(prefix_path)

    # make dir 'best_model'
    save_best_model = 'best-model' # train 중 가장 좋은 모델의 weight 정보 저장
    save_model_path = home + save_best_model + '/'

    if not os.path.isdir(home + save_best_model):
        try:
            os.mkdir(home + save_best_model)
        except:
            pass

    ## 모델별 path 구성
    json_path_model = 'model-info-json'     # 기본적인 모형 관련 정보가 저장되는 json_path(ML/DL 공통)
    json_path_visual = 'for-visual-json'    # classification일 경우, json_path 하나 더 생성됨
    json_path_shap = 'shap-value-json'      # shap_value(feature_contribution) json 파일 저장하는 path(ML/DL 공통)

    path_list = list()
    if 'keras' in get.model_type(model_name) or 'tf' in get.model_type(model_name) or 'pytorch' in get.model_type(model_name):
        csv_path_visual = 'for-visual-csv'            # learning history가 저장되는 csv_path

        path_list = [json_path_model, csv_path_visual, json_path_visual, json_path_shap]   # DL

    elif 'lightgbm' in get.model_type(model_name):
        metric_lightgbm = get.from_model(model_name).all_params()['metric']
        if metric_lightgbm in metricsLightgbm.CLASSIFICATION:
            path_list = [json_path_model, json_path_visual, json_path_shap]       # lightgbm & Classification
        elif metric_lightgbm in metricsLightgbm.REGRESSION:
            path_list = [json_path_model, json_path_shap]                         # lightgbm & Regression
        else:
            pass
        
    elif 'Classifier' in get.model_type(model_name) or 'Logistic' in get.model_type(model_name) or 'SVC' in get.model_type(model_name) or 'NB' in get.model_type(model_name):
        path_list = [json_path_model, json_path_visual, json_path_shap]       # ML & Classification

    else:
        path_list = [json_path_model, json_path_shap]                         # ML & Regression
    
    for j in [i for i in path_list if not os.path.isdir(i)]:
        try:
            os.mkdir(j)
        except FileExistsError:
            pass

    
    ## 모델별 trial_num 생성
    if 'csv_path_visual' in locals():
        trial_num_list = list([get.trial_number_all(i, 'json') for i in path_list if 'json' in i])
        trial_num_list.append([get.trial_number_all(i, 'csv') for i in path_list if 'csv' in i])
        merged_list = list(itertools.chain.from_iterable(trial_num_list))
        merged_list = list(map(int, merged_list))
        trial_num = str(max(merged_list))

    else:
        trial_num_list = [get.trial_number_all(i, 'json') for i in path_list if 'json' in i]
        trial_num = str(max(trial_num_list))
    
    ## 최종 path
    if len(path_list) == 2:     # ML & Regression
        json_file_name_model = json_path_model + '/' + get.model_type(model_name) + '-' + trial_num + '.json'
        json_file_name_shap = json_path_shap + '/' + get.model_type(model_name) + '-shap-' + trial_num + '.json'
        dir_dict = {'home_path': home,
                    'prefix_path': prefix_path,
                    'save_model_dir': save_best_model,
                    'save_model_joblib': save_model_path,
                    'model_json': json_file_name_model,
                    'shap_json': json_file_name_shap,
                    'model_json_full': home + prefix_path + json_file_name_model,
                    'shap_json_full': home + prefix_path + json_file_name_shap}

    elif 'json' in path_list[1]: # ML & Classification
        json_file_name_model = json_path_model + '/' + get.model_type(model_name) + '-' + trial_num + '.json'
        json_file_name_visual = json_path_visual + '/' + get.model_type(model_name) + '-visual-' + trial_num + '.json'
        json_file_name_shap = json_path_shap + '/' + get.model_type(model_name) + '-shap-' + trial_num + '.json'

        dir_dict = {'home_path': home,
                    'prefix_path': prefix_path,
                    'save_model_dir': save_best_model,
                    'save_model_joblib': save_model_path,
                    'model_json': json_file_name_model,
                    'visual_json': json_file_name_visual,
                    'shap_json': json_file_name_shap,
                    'model_json_full': home + prefix_path + json_file_name_model,
                    'visual_json_full': home + prefix_path + json_file_name_visual,
                    'shap_json_full': home + prefix_path + json_file_name_shap}

    elif len(path_list) == 4: # DL
        csv_file_name_visual = csv_path_visual + '/' + get.model_type(model_name) + '-history-' + trial_num + '.csv'
        json_file_name_model = json_path_model + '/' + get.model_type(model_name) + '-' + get.trial_num_json(csv_path_visual, 'csv') + '.json'
        json_file_name_visual = json_path_visual + '/' + get.model_type(model_name) + '-visual-' + get.trial_num_json(csv_path_visual, 'csv') + '.json'
        json_file_name_shap = json_path_shap + '/' + get.model_type(model_name) + '-shap-' + get.trial_num_json(csv_path_visual, 'csv') + '.json'

#         save_model_hdf5 = home + prefix_path + save_best_model + '/'  + 'trial-' + get.trial_num_json(csv_path_visual, 'csv')
#         save_model_hdf5 = home + save_best_model + '/'

        dir_dict = {'home_path': home,
                    'prefix_path': prefix_path,
                    'save_model_dir': save_best_model,
                    'save_model_path': save_model_path,
                    'model_json': json_file_name_model,
                    'visual_csv': csv_file_name_visual,
                    'visual_json': json_file_name_visual,
                    'shap_json': json_file_name_shap,
                    'model_json_full': home + prefix_path + json_file_name_model,
                    'visual_csv_full': home + prefix_path + csv_file_name_visual,
                    'visual_json_full': home + prefix_path + json_file_name_visual,
                    'shap_json_full': home + prefix_path + json_file_name_shap}

    else:
        dir_dict = dict()

    # 기존 path로 변경
    os.chdir(original_path)
    
    return dir_dict


